from IPython.display import display, HTML
import uuid


class Progress:
    """
    Add a progressbar.
    Using https://getbootstrap.com/docs/4.3/components/progress/ not the HTML5 one.
    Note that Jupyter uses an older version: 3.4 (https://getbootstrap.com/docs/3.4/components/), but I doubt this will be so for long hence the color via style.
    Two uses different uses.

    Timer:
    >>> Progress().countdown(seconds) = the progressbar counts to n seconds every tick ms

    Manual
    >>> p = Progress(percent=20, label='hello') both percent and label are optional arguments
    >>> p.percent = 30
    >>> p.label = 'bye'
    Do note that `.percent` and `.label` do not check what the real value of the Progress bar is and rely on a previously stored value.
    Hence why they are incorrect for `.countdown()`.
    This is unavoidable as JS can run a IPython command only when the kernel is idle and this is needed to input values.
    Consequently, the method `.update('varname')` is provided which will check.
    The argument for .update is the name variable has in the Python mainspace as a string.
    Do note the update happens after the cell has finished excecuting. So
    >>> p = Progress(percent=20, label='hello')
    >>> print(p.percent, p.label) # correct
    >>> p.countdown(5)
    >>> time.sleep(1)
    >>> print(p.percent, p.label) # incorrect
    >>> p.update('p')
    >>> print(p.percent, p.label) # incorrect within the same cell, correct in next.
    The HTML element id of the bar is `.id`.
    """

    def __init__(self, percent=0, label=''):
        self.id = uuid.uuid1().hex
        self._percent = percent
        self._label = label
        display(HTML(f'''<div class="progress">
          <div id="{self.id}"
              class="progress-bar progress-bar-striped" role="progressbar"
              style="background-color: #ffc107; width: {self._percent}%;">
              {self._label}
          </div>
        </div>
        '''))

    def countdown(self, seconds, tick=1000):
        self._append(code=f'''
            window.timegone = {seconds * self.percent / 100}; // zero normally...
            window.killTimer = setInterval(() => {{
              if(++window.timegone > {seconds}) {{
                  clearInterval(killTimer);
                  $('#{self.id}').css('width', '100%')
                                 .removeClass('progress-bar-striped')
                                 .css('background-color','#28a745')
                                 .html('Complete');
              }} else {{$('#{self.id}').css('width', (window.timegone/{seconds}*100)+'%')}}
            }}, {tick});''')
        return self

    def _set_percent(self, percent):
        ## sets the progress to that percent
        self._percent = percent
        if percent < 100:
            self._append(code=f"$('#{self.id}').css('width', '{percent}%');")
        else:
            self._append(
                code=f"$('#{self.id}').css('width', '100%').css('background-color','#28a745').removeClass('progress-bar-striped');")
        return self

    def _set_label(self, label):
        # add a label
        self._label = label
        self._append(code=f"$('#{self.id}').html('{label}');")
        return self

    def update(self, name):
        ## name is a string with the name in the main namespace that the progress instance is called by
        self._append(
            code=f'''IPython.notebook.kernel.execute("{name}._label  = '"+$('#{self.id}').html().replace('\\n',' ')+"'");''')
        self._append(
            code=f'''IPython.notebook.kernel.execute("{name}._percent  = "+$('#{self.id}').width()/$('#{self.id}').parent().width() * 100);''')

    label = property(lambda self: self._label, _set_label)
    percent = property(lambda self: self._percent, _set_percent)

    def _unbloat(self, n):
        # a wee hack to stop newlines appearing.
        return f'$("#{n}").parent().parent().hide();'

    def _append(self, code, hidden=True):
        # adds JS
        n = uuid.uuid1().hex
        if hidden:
            code += '\n' + self._unbloat(n)
        display(HTML(f'''<script type="text/javascript" id='{n}'>
        {code}
        </script>'''))




from typing import *
# if TYPE_CHECKING:   # from typing
#     import pandas as pd

from rdkit import Chem
import os, json
from typing import List, Sequence
from warnings import warn

class TableMixin:

    def pandas_to_mols(self,
                      df: 'pd.DataFrame',
                      targetfolder: str,
                      name_column_name: str = 'name',
                      mol_column_name: str = 'mol',
                      skip_first=False) -> List[str]:
        """
        Extracts the mols


        :param df: pandas datatable. Most likely the dataframe is ``rdkit.Chem.PandasTools`` flavoured,
            but this is not a requirement
        :param targetfolder: folder where to save the mol files (in a github repo)
        :param name_column_name: column name containing the molecule name
        :param mol_column_name: column name containing the molecules
        :param skip_first: default False, skip first if its definitions XChem style.
        :return:
        """
        filenames = []
        for i, row in df.iterrows():
            if skip_first and i == 0:
                continue
            mol = row[mol_column_name]
            name = row[name_column_name]
            if mol is None:
                filenames.append(None)
                continue
            elif not name:
                filename = f'compound_{i}'
            else:
                def valid(character): any([character.isalpha(),
                                           character.isdigit(),
                                           character in (' ', '-', '_', '.')
                                           ])
                filename = ''.join(filter(valid, name)).strip()
            filename += '.mol'
            Chem.MolToMolFile(mol, filename)
            filenames.append(filename)
        return filenames

    def sdf_to_mols(self, sdfile: str, targetfolder: str, skip_first=False) -> List[str]:
        """
        Converts a sdf file to mols for github
        The format is the same as that required by XChem fragalysis

        :param sdfile: SDF file
        :param targetfolder: folder where to save the mol files (a github repo)
        :param skip_first: default True, skip first if its definitions XChem style.
        :return: list of filenames
        """
        if os.path.exists(targetfolder):
            warn(f'{targetfolder} already exists')
        else:
            os.mkdir(targetfolder)
        suppl = Chem.SDMolSupplier(sdfile)
        next(suppl)
        filenames = []
        for i, mol in enumerate(suppl):
            if i == 0 and skip_first:
                continue
            filename = os.path.join(targetfolder, mol.GetProp('_Name') + '.mol')
            Chem.MolToMolFile(mol, filename)
            filenames.append(filename)
        return filenames

    def sdf_to_meta(self, sdfile: str) -> dict:
        suppl = Chem.SDMolSupplier(sdfile)
        return next(suppl).GetPropsAsDict()

    def sdf_to_json(self,
                    sdfile: str,
                    keys: Sequence[str],
                    key_defaults: Sequence[float],
                    filename: str,
                    spaced:bool=True
                    ) -> List[list]:
        """
        Get the keys out of the sdf and make a list of lists. filling blanks with the defaults.
        A convoluted method, that evolved into doing what numpy does...

        :param sdfile: SDF file
        :param keys: props to show in the table
        :param key_defaults: what are the value to show if absent (e.g. ∆∆G of 999.)
        :param filename: with path data.json
        :param spaced: convert underscores to spaces in header?
        :return:
        """
        suppl = Chem.SDMolSupplier(sdfile)
        # Parse data
        data = []
        for i, mol in enumerate(suppl):
            if i == 0:
                continue
            d = {'name': mol.GetProp('_Name')}
            for k, kd in zip(keys, key_defaults):
                if mol.HasProp(k):
                    v = mol.GetProp(k).strip()
                    if k == 'ref_mols':
                        d[k] = v.replace(',', ', ')
                        continue
                    elif isinstance(kd, float):
                        try:
                            v = round(float(v) * 10) / 10
                        except:
                            v = kd
                    elif isinstance(kd, str):
                        pass
                    d[k] = v
                else:
                    d[k] = kd
            data.append(d)
        # Make header
        if not spaced:
            header = [['name']+list(keys)]
        else:
            header = [[k.replace('_',' ') for k in ['name']+list(keys)]]
        # Combine
        flatten = header + [list(d.values()) for d in data]
        json.dump(flatten, open(filename, 'w'))
        return flatten

    def make_fragment_table(self,
                   metadata:Dict[str, str],
                   username:str,
                   repo_name:str,
                   foldername:str,
                   protein_sele:str,
                   sort_col:int,
                   sort_dir:str='asc',
                   template_row:int=-1,
                   fragment_row:int=-1,
                   jsonfile:str='data.json',
                   branch:str='main'):
        """
        Makes a interactive table out of xchem submission.
        NB. this formerly accepted an SDF (XChem style formatted) not it is agnostic
        and accepts the dictionary metadata.

        :param metadata: a dictionary of value -> definitions. It is for the ``description`` text and
            not for the table per se. If really must be ommitted simply pass an empty dictionary.
            also see ``sdf_to_meta(sdfile)``
        :param username: GitHub username
        :param repo_name: GitHub repo name
        :param foldername: folder name within repo
        :param protein_sele: selection on protein to show as sticks
        :param sort_col: column to sort by
        :param sort_dir: asc | desc
        :param template_row: -1 for foldername/template.pdb else row in data.json to load.
        :param fragment_row: -1 for none. else read that column. comma separated. Names must match mol in foldername.
        :param jsonfile: data. use ``sdf_to_json``.
        :param branch: main or master etc.
        :return:
        """
        giturl = f'https://raw.githubusercontent.com/{username}/{repo_name}/{branch}/{foldername}'
        # Description
        self.description += '## Fields\n\n' + ''.join([f'* **{k}**: {v}\n' for k, v in metadata.items()]) + '\n'
        self.description += '## Data\n'
        self.description += '<table id="data" class="display" width="100%"></table>\n'
        self.description += '<link href="https://cdn.datatables.net/1.10.20/css/dataTables.bootstrap4.min.css" rel="stylesheet"/>\n'
        # JS
        self.loadfun += '''["https://cdn.datatables.net/1.10.20/js/jquery.dataTables.min.js",
                            "https://cdn.datatables.net/1.10.20/js/dataTables.bootstrap4.min.js"].map((v,i) => {
                                setTimeout(() => {
                                    let s = document.createElement("script");
                                    s.type = "text/javascript";
                                    s.src = v;
                                    $("body").append(s);
                                },100 * i + 1);
                            });'''
        load_table = f'const giturl = "{giturl}"; const jsonfile = "{jsonfile}";' + \
                     f'const sort_col = {sort_col}; const sort_dir= "{sort_dir}";' + \
                     f'const protein_sele = "{protein_sele}"; const template_row={template_row};' +\
                     f'const fragment_row={fragment_row};' +\
                     '''fetch(`${giturl}/${jsonfile}`)
                        .then((response) => {
                            return response.json();
                        })
                        .then((dataSet) => {
                            let headers = dataSet.shift().map(v => ({'title': v}));
                            window.dataSet = dataSet;
                            window.dt = $('#data').DataTable({
                                    data: dataSet,
                                    columns:  headers,
                                    scrollX: true,
                                    order: [[ sort_col, sort_dir ]]
                                });
                            window.makeTableClickable(giturl, protein_sele, template_row, fragment_row);
                        });\n'''
        self.loadfun +=  '''window.makeTableClickable = (giturl, protein_sele, template_row, fragment_row) => {
                        $('#data tbody').css('cursor', 'pointer');
                        $('#data tbody').on('click', 'tr', function () {
                        $('#data .bg-info').removeClass('bg-info');
                        $(this).addClass('bg-info');
                        let data = dt.row(this).data();
                        let name = data[0].trim();
                        let hits = (fragment_row === -1) ? [] : data[fragment_row].replace(' ', '').replace(/_0/g, '').split(',');
                        if (!myData.proteins.some(v => v.name === name)) {
                            let ref_url = giturl+'/';
                            ref_url += (template_row === -1) ? 'template.pdb' : data[template_row].replace('.pdb','')+'.pdb';
                            myData.proteins.push({
                                type: 'url',
                                value: ref_url,
                                'name': name,
                                'loadFx': (protein) => {
                                    let colors = ['silver', 'black', 'white', 'gray', 'gainsboro'];
                                    protein.removeAllRepresentations();
                                    protein.addRepresentation("cartoon", {opacity: 0.1, colorValue: 'coral'});
                                    protein.addRepresentation("hyperball", {sele: protein_sele, colorValue: 'coral', opacity: 0.3});
                                    protein.autoView(protein_sele, 2000);
                                    // ligand
                                    NGL.getStage('viewport').loadFile(
                                        `${giturl}/${data[0].replace('.mol','')}.mol`)
                                        .then(molecule => {
                                            molecule.removeAllRepresentations();
                                            molecule.addRepresentation("hyperball", {sele: 'not _H', colorValue: 'teal', opacity: 0.3});
                                            molecule.autoView(2000);
                                        });
                    
                                    hits.map(h => NGL.getStage('viewport').loadFile(
                                        `${giturl}/${h.trim()}.mol`)
                                        .then(molecule => {
                                            molecule.removeAllRepresentations();
                                            molecule.addRepresentation("licorice", {
                                                sele: 'not _H',
                                                colorValue: colors.pop(),
                                                multipleBond: "symmetric"
                                            });
                                        })
                                    );
                                }
                            });
                        }
                        NGL.specialOps.load(name, false);
                    });};'''
        self.loadfun += '$(document).ready(() => { setTimeout(() => {' +load_table +'},200)});'


from typing import *
import json, re

class Prolink:
    """
    Create prolink and export it with str(prolink).
    Allows for handy manupulation such as copying and changing its values.

    To instantiate from string tag, use ``Prolink.parse``.
    """
    def __init__(self,
                 text:str,
                 target:str = "viewport",
                 focus:str="domain",
                 selection: str="*",
                 color: Optional[str]=None, #"green",
                 radius: Optional[int]=None, #=5,
                 tolerance: Optional[int]=None, #1,
                 hetero: Optional[bool]=None, #False,
                 label: Optional[bool]=None, #True,
                 view: Optional[str] = None,
                 title: Optional[str]=None,
                 load:Optional[str]=None,
                 alts: Optional[List[Dict[str, str]]] = None):
        """

        The arguments are data tags.
        For full details of these see https://michelanglo.sgc.ox.ac.uk/docs/markup

        >>> pro = Prolink(text='V242A', focus="domain", color="#967bb6", selection=":V", alts=[dict(selection="242:V",focus="residue", color="#FF00FF")])
        >>> str(pro)
        >>> pro.color = 'yellow'

        The alternative selection differs (``alts`` as opposed separate fields).
        As a result of alts remember to do `copy.deepcopy` and not just `copy.copy` if copying.

        >>> import copy
        >>> new_pro = copy.deepcopy(pro)

        NB. data-toggle="protein" defines a prolink, so isnt an option.
        NB2. intolerant to spelling "tollerance [sic.]"

        :param text: inside text, assumed to be escaped already
        :param target: viewport id in JS (generally "viewport")
        :param focus: residue | domain | residue | clash | bfactor | surface | overlay * | domain-overlay *
        :param selection: see NGL selection
        :param color: for now a string in a valid color (``#ff00ff`` or ``green``)
        :param radius:
        :param tolerance:
        :param hetero:
        :param label:
        :param view:
        :param title: caption
        :param load:
        :param alts: ``alts`` is a list of dict with keys ``selection`` | ``focus`` | ``color``. e.g. [{'selection': '242:V', 'focus': 'residue' color: "#FF00FF"}]
        :type alts: List[Dict[str, str]]
        """
        self.element = 'span'
        self.text = text
        self.target = target
        self.focus = focus
        self.selection = selection
        self.color = color
        self.title = title
        self.load = load
        self.radius = radius
        self.tolerance = tolerance
        self.hetero = hetero
        self.label = label
        self.view = view
        self.alts = alts if alts else []

    @classmethod
    def parse(cls, tag: str):
        """
        Parse a tag (str) to create a Prolink instance
        Round trip:

        >>> pro = Prolink(text='Hello', focus="domain", color="pink", selection=":A")
        >>> spro = str(pro)
        >>> neopro = Prolink.parse(spro)
        >>> print(pro, neopro)

        :param tag: tag contain both opening and closing elements. Can contain newlines
        :return: Prolink instance
        """
        # split
        rex = re.match('<(\w+) ([\w\W]*)>([\w\W]*)<\/(\w+)\s?>', tag.strip())
        assert rex is not None, 'Is the tag closed?'
        fore_element, raw_attributes, text, aft_element = rex.groups()
        assert fore_element == aft_element, f'The closing element differs: {fore_element} != {aft_element}'
        esc_attributes = re.findall('data-([\w\-]+)\=((?:(?!data-).)*)', raw_attributes)
        # escape and collapse alts
        attributes = {'alts': []}
        for k, v in esc_attributes:
            if k == 'toggle':
                continue
            try:
                value = json.loads(v)
            except: # its a string already? html is very tolerant...
                value = v
            if '-alt' not in k:
                attributes[k] = value
            else:
                ak, raw_ai = re.match('(\w+)-alt(\d+)', k).groups()
                ai = int(raw_ai) - 1
                while len(attributes['alts']) <= ai:
                    attributes['alts'].append({})
                attributes['alts'][ai][ak] = value
        # instantiate
        self = cls(text=text, **attributes)
        self.element = fore_element
        return self


    @property
    def data_attributes(self):

        attributes = {'data-toggle': "protein",
                        'data-target': self.target,
                        'data-focus': self.focus,
                        'data-selection': self.selection,
                      }
        for optional in ('load', 'title', 'color', 'tolerance', 'radius', 'hetero', 'label', 'view'):
            if getattr(self, optional) is not None:
                attributes[f'data-{optional}'] = getattr(self, optional)
        for i, alt in enumerate(self.alts):
            for k, v in alt.items():
                attributes[f'data-{k}-alt{i+1}'] = v
        return attributes

    def escaper(self, value):
        return json.dumps(value)

    def __str__(self):
        data = ' '.join([f'{k}={self.escaper(v)}' for k, v in self.data_attributes.items()])
        return f'<{self.element} class="prolink" {data}>{self.text}</{self.element}>'



from typing import Optional, List, Dict, Tuple

import re

def deindent(text: str) -> str:
    return re.sub('\s+', ' ', text)


def make_modal(title: str, body: str, id_attribute: str) -> str:
    modalblock = f'''<div class="modal fade" 
                          id="{id_attribute}" 
                          data-backdrop="static" data-keyboard="false" 
                          tabindex="-1" 
                          aria-labelledby="staticBackdropLabel" aria-hidden="true">
      <div class="modal-dialog modal-dialog-centered modal-xl">
        <div class="modal-content">
          <div class="modal-header">
            <h5 class="modal-title" id="staticBackdropLabel">{title}</h5>
            <button type="button" class="close" data-dismiss="modal" aria-label="Close">
              <span aria-hidden="true">&times;</span>
            </button>
          </div>
          <div class="modal-body">
            {body}
          </div>
          <div class="modal-footer">
            <button type="button" class="btn btn-secondary" data-dismiss="modal">Close</button>
          </div>
        </div>
      </div>
    </div>
    '''
    return modalblock


def make_carousel(bodies: list, id_attribute: str)  -> str:
    assert len(bodies) >= 2, 'You need two for a carousel'

    def inner(text, active=False):
        return f'''<div class="carousel-item {'active' if active else ''}" data-interval="false">
          <div class="w-100 h-100">
              {text}
          </div>
        </div>'''

    li_els = [f'<li class="bg-dark active" data-target="#{id_attribute}" data-slide-to="0"></li>'] + \
             [f'<li class="bg-dark" data-target="#{id_attribute}" data-slide-to="{i}"></li>' for i in
              range(1, len(bodies))]
    li_elblock = '\n'.join(li_els)
    div_els = [inner(bodies[0], active=True)] + [inner(bodies[i], active=False) for i in range(1, len(bodies))]
    div_elblock = '\n'.join(div_els)
    carouselblock = f'''
    <div id="{id_attribute}" class="carousel slide" data-interval="false" data-ride="carousel">
      <ol class="carousel-indicators">
        {li_elblock}
      </ol>
      <div class="carousel-inner" style="min-height: 800px;">
        {div_elblock}
      </div>
      <a class="carousel-control-prev text-dark" href="#{id_attribute}" role="button" data-slide="prev">
        <i class="fas fa-caret-left  fa-2x text-dark"></i>
        <span class="sr-only">Previous</span>
      </a>
      <a class="carousel-control-next  text-dark" href="#{id_attribute}" role="button" data-slide="next">
        <i class="fas fa-caret-right  fa-2x text-dark"></i>
        <span class="sr-only">Next</span>
      </a>
    </div>
    '''
    return carouselblock


def make_list(entries: list, id_attribute: Optional[str] = None, ordered=False) -> str:
    lis = ' '.join([f'<li class="list-group-item">{entry}</li>' for entry in entries])
    if ordered:
        el = 'ol'
    else:
        el = 'ul'
    if id_attribute:
        ol = f'''<{el} class="list-group" id="{id_attribute}">{lis}</{el}>'''
    else:
        ol = f'''<{el} class="list-group">{lis}</{el}>'''
    return ol


def make_FA_list(entries: List[Tuple[str, str]], id_attribute: Optional[str] = None) -> str:
    lis = ' '.join([f'<li><span class="fa-li"><i class="{fa}"></i></span>{entry}</li>' for fa, entry in entries])
    if id_attribute:
        ul = f'''<ul class="fa-ul" id="{id_attribute}">{lis}</ul>'''
    else:
        ul = f'''<ul class="fa-ul">{lis}</ul>'''
    return ul


def make_slide(title: str, body: str) -> str:
    return f'''<div class="p-5">
<h1>{title}</h1>

<br/>

    <div style="font-size: 1.5em;">
        {body}
    </div>

</div>
'''


def make_header_slide(title: str, subtitle: str) -> str:
    if subtitle:
        wrapped_sub = f'<hr class="my-4"/><p class="lead">{subtitle}</p>'
    else:
        wrapped_sub = ''
    return f'<div class="jumbotron p-5"><h1 class="display-4">{title}</h1>{wrapped_sub}</div>'



import requests, json, pickle, re, os
from warnings import warn
from typing import Dict, Optional
from datetime import datetime, timedelta
import time

from .page import MikePage
from .enums import Privacy, Location

try:
    from .progressbar import Progress
except:
    Progress = None


class MikeAPI:
    Privacy = Privacy
    Location = Location

    def __init__(self,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 session: Optional[requests.Session] = None,
                 url: str = 'https://michelanglo.sgc.ox.ac.uk/'):
        """
        Gets the API interface thingamabob.

        :param username: if None, check environment variable ``MICHELANGLO_USERNAME`` or prompt
        :param password: if None, check environment variable ``MICHELANGLO_PASSWORD`` or prompt
        :param session: supply your own session, debug use basically.
        :param url: specify if using anything other than sgc version (e.g. 'http://0.0.0.0:8088')
        """
        if url[-1] != '/':
            url = url + '/'
        if '://' not in url:
            url = 'https://' + url
        self.url = url
        self.username = self._retrieve_arg(username, 'MICHELANGLO_USERNAME')
        self.password = self._retrieve_arg(password, 'MICHELANGLO_PASSWORD')
        self.visited_pages = []  # filled by self.refresh_pages()
        self.owned_pages = []  # filled by self.refresh_pages()
        self.public_pages = []  # filled by self.refresh_pages()
        if session:
            self.request = session
        else:
            self.request = requests.Session()
        self.login()
        self.refresh_pages()

    def _retrieve_arg(self, arg, envname):
        """
        fill username/password.

        :param arg:
        :param envname: 'MICHELANGLO_PASSWORD' or 'MICHELANGLO_USERNAME'
        :return:
        """
        if arg is not None:
            return arg
        elif envname in os.environ:
            return os.environ[envname]
        else:
            return input(envname).strip()

    # ==================================================================================================================

    def login(self):
        reply = self.post('login', data={'username': self.username,
                                         'password': self.password,
                                         'action': 'login'})
        if reply.status_code != 200:
            raise ConnectionRefusedError(f'Code {reply.status_code}: {reply.content}')
        return self

    def verify_user(self):
        return self.post_json('login', {'action': 'whoami'})

    # ==================================================================================================================

    def post(self, route, data=None, headers=None):
        reply = self.request.post(self.url + route, data, headers)
        if reply.status_code == 200:
            return reply
        else:
            raise Exception('The site {url} returned a status code {code}. Content: {con}' \
                            .format(url=self.url, code=reply.status_code, con=reply.content))

    def post_json(self, route, data=None, headers=None):
        return self.post(route, data, headers).json()

    # ==================================================================================================================

    def refresh_pages(self):
        data = self.post_json('get_pages')
        self.visited_pages = [MikePage(self, uuid) for uuid in data['visited']]
        self.owned_pages = [MikePage(self, uuid) for uuid in data['owned']]
        self.public_pages = [MikePage(self, uuid) for uuid in data['public']]
        self.all_pages = [MikePage(self, uuid) for uuid in data['all']]  # admin only
        return self

    def get_page(self, uuid: str) -> MikePage:
        uuid = self._clean_uuid(uuid)
        try:
            data = self.post_json('data/' + uuid, data={'mode': 'json'})
            return MikePage(self, uuid).parse(data)
        except Exception as err:
            raise err

    def _clean_uuid(self, url):
        if isinstance(url, str):
            return url.split('/')[-1]
        elif isinstance(url, dict):
            return url['page']

    def set_page(self, uuid, page):
        uuid = self._clean_uuid(uuid)
        data = page.dumps()
        return self.post_json('edit_user-page', data=data)

    def del_page(self, uuid):
        uuid = self._clean_uuid(uuid)
        return self.post_json('delete_user-page', data={'page': uuid})

    def rename_page(self, uuid, new_name):  ##admin only
        warn('This is depracated use shorten')
        uuid = self._clean_uuid(uuid)
        return self.post_json('rename_user-page', data={'old_page': uuid, 'new_page': new_name})

    def shorten_page(self, uuid, short_name):
        return self.post_json('set', {'item': 'shorten',
                                      'short': short_name.lower(),
                                      'long': uuid})

    # ==================================================================================================================

    def convert_pdb(self, code=None, filename=None, **prolink_settings):
        """
        use underscores for the hyphens in the data attr!
        >>> new_page = mike.convert_pdb(code='1UBQ', data_focus='residue', data_selection='20:A')
        >>> new_page = mike.convert_pdb(filename='/home/my_protein.pdb')
        >>> mike.page_link(new_page['page'])
        """
        data = {}
        prolink_settings['role'] = 'NGL'
        if code:
            prolink_settings['data-load'] = code
            data['mode'] = 'code'
            data['pdb'] = code
        elif filename:
            data['mode'] = 'file'
            data['pdb'] = open(filename).read()
        else:
            raise ValueError('Specifiy at least a pdb `code` or a `file`')
        data['viewcode'] = ' '.join(
            ['{a}="{p}"'.format(p=prolink_settings[s], a=s.replace('_', '-')) for s in prolink_settings])
        reply = self.post_json('convert_pdb', data)
        return MikePage(self, reply['page'])

    # ==================================================================================================================

    def convert_pse(self, demo_filename=None, filename=None, sticks='hyperball', uniform_non_carbon=False,
                    pdb_code=None):
        """
        filename is local while demo_filename is on server.
        """
        data = {'stick_format': sticks, 'uniform_non_carbon': uniform_non_carbon}
        if demo_filename:
            if demo_filename.find('.pse') == -1:
                warn(f'{demo_filename} does not end in .pse. Forcing.')
                demo_filename += '.pse'
            data['demo_filename'] = demo_filename
        elif filename:
            data['file'] = open(filename).read()
            if pdb_code:
                data['pdb'] = pdb_code
            else:
                data['pdb'] = False
        else:
            raise ValueError('Demo or file')
        reply = self.post_json('convert_pse', data)
        return MikePage(self, reply['page'])

    @staticmethod
    def print_reply(reply):
        print(f'Status code: {reply.status_code}; Headers: {reply.headers}; Content: {reply.content}')

    # def rename_pdb(self, data, name: str, idx: Optional[int] = None, original: Optional[str] = None):
    #     # original is the name in the original version. idx is the index
    #     if original is not None:
    #         for p in data['pdb']:
    #             if p[0] == original:
    #                 p[0] = name
    #                 break
    #         for p in data['proteinJSON']:
    #             if p['value'] == original:
    #                 p['value'] = name
    #                 break
    #     elif idx is not None:
    #         data['proteinJSON'][idx]['value'] = name
    #         data['pdb'][idx][0] = name
    #     else:
    #         raise TypeError('specify either idx or original')

    # ======================================================================================================================

    def set_toast(self, title: str, description: str, bg: str = 'bg-danger'):
        # admin only.
        return self.post_json('set', {'item': 'msg',
                                      'title': title,
                                      'descr': description,
                                      'bg': bg})

    def reset(self, change: str = 'misc.', timeout: int = 300):
        """
        Reset the server after ``timeout`` seconds. warning about the reset with a custom message ``change``.

        :param change:
        :param timeout:
        :return:
        """
        assert 'MIKE_SECRET' in os.environ, 'Please provide the secret code as an environment variable.'
        self.set_toast(title='<i class="far fa-danger"></i> The server reset',
                       description=f'In order to implement the latest changes ({change}), the server will reset at ' + \
                                   f'{datetime.now() + timedelta(seconds=timeout)} BST ({timeout} sec. countdown). ' + \
                                   f'This will be a brief blip. You might not even notice it!',
                       bg='bg-danger')
        if Progress is not None:
            Progress().countdown(timeout)
        time.sleep(timeout)
        try:
            print(self.post_json('set', {'item': 'terminate', 'code': os.environ['MIKE_SECRET']}))
        except:
            print('Server resetting.')

    def get_user_email(self, username: str):
        assert 'MIKE_SECRET' in os.environ, 'Please provide the secret code as an environment variable.'
        reply = self.post_json('login', {'action': 'email', 'username': username, 'code': os.environ['MIKE_SECRET']})
        if 'email' not in reply:
            raise ValueError(str(reply))
        return reply['email']



########################################################################################################################
__doc__ = \
    """
    Pip install file. All the files have this block here, so in PyCharm I can change them en-mass.
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "10 July 2020 A.D."
__license__ = "MIT"
__version__ = "1.1.2"
__citation__ = "None."

########################################################################################################################


from setuptools import setup, find_packages

import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='michelanglo_api',
    version= __version__,
    packages=find_packages(),
    url='https://github.com/matteoferla/MichelaNGLo-api',
    license=__license__,
    author=__author__,
    author_email=__email__ ,
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='API to interface with Michelanglo'
)



try:
    from IPython.display import display, HTML
except ImportError:
    pass  # not jupyter.

import json, re, os, pickle
from warnings import warn
from typing import Any
from datetime import datetime
from typing import Optional, Dict, List

from .enums import Privacy, Location

try:
    from .table import TableMixin
except ImportError:
    class TableMixin:
        pass


class MikePage(TableMixin):
    """
    This class holds page data.
    And adds some extra conversion features.
    """
    parent = None

    preferences = {'authors': list,
                   'backgroundcolor': str,
                   'columns_text': int,
                   'columns_viewport': int,
                   'confidential': bool,
                   'current_page': str,
                   'data_other': str,
                   'date': datetime,
                   'descr_mdowned': str,
                   'description': str,
                   'editable': bool,
                   'editors': list,
                   'encrypted': bool,
                   'encryption': bool,
                   'encryption_key': (None.__class__, str),
                   'firsttime': bool,
                   'freelyeditable': bool,
                   'image': (None.__class__, bool, str),  #: todo fix! this legacy business
                   'is_unseen': bool,
                   'key': (None.__class__, str),
                   'loadfun': str,
                   'location_viewport': Location,
                   'model': bool,
                   'no_analytics': bool,
                   'no_buttons': bool,
                   'no_user': bool,
                   'page': str,
                   'pdb': dict,  # not list of tuples!
                   'proteinJSON': list,
                   'public': Privacy,
                   'revisions': list,
                   'save': bool,
                   'stick': str,
                   'stick_format': str,
                   'title': str,
                   'uniform_non_carbon': bool,
                   'user': str,
                   'validation': (bool, None.__class__),
                   'verbose': bool,
                   'viewport': str,
                   'visitors': list}

    definitions = {'authors': 'List of who has edited the page',
                   'backgroundcolor': '(read-only) colour of background of NGL viewport',
                   'columns_text': 'x/12 width of description',
                   'columns_viewport': 'x/12 width of NGL viewport',
                   'confidential': 'show confidential banner?',
                   'current_page': '(read-only) uuid as present in the data (not redirect)',
                   'data_other': 'NGL viewport commands (initial view in the absence of loadfun',
                   'date': '(read-only) last access (see page delition policy)',
                   'descr_mdowned': '(read-only) description in HTML',
                   'description': 'description in MD',
                   'editable': '(read-only) Legacy',
                   'editors': 'Who has been added as a potential editor',
                   'encrypted': '(read-only) Encrypt? (DB)',
                   'encryption': 'Encrypt?',
                   'encryption_key': 'Encryption key. Not stored in the DB',
                   'firsttime': '(read-only) Show first time viewing notifications? Do not recall dif from is_unseen',
                   'freelyeditable': 'Can anyone edit?',
                   'image': 'Image link to show or None',
                   'is_unseen': '?? Same as firsttime??',
                   'key': '?? Same as encryption key??',
                   'loadfun': 'JS to add',
                   'location_viewport': 'Left or Right NGL viewport',
                   'model': 'Is this a model?',
                   'no_analytics': '(read-only) Do not gatether stats on this page?',
                   'no_buttons': '(read-only) Do not show buttons on this page?',
                   'no_user': '(read-only) Do not show user controls on this page?',
                   'page': '(read-only) uuid',
                   'pdb': 'list of tuple of name and block',
                   'proteinJSON': 'list of protein',
                   'public': Privacy,
                   'revisions': 'list of past edits',
                   'save': '????',
                   'stick': '(ready_only) Legacy',
                   'stick_format': '(ready_only) Legacy',
                   'title': 'Title',
                   'uniform_non_carbon': '(ready_only) Legacy',
                   'user': '(ready_only) You',
                   'validation': '(ready_only) Legacy',
                   'verbose': '(ready_only) Legacy',
                   'viewport': '(ready_only) NGL viewport name',
                   'visitors': '(ready_only) who visited'}

    def __init__(self, parent, uuid: str):
        """
        Parent is MikeAPI

        :param parent:
        :param uuid:
        """
        self.parent = parent
        self.page = uuid
        self.date = datetime.min
        self.proteins = []
        self.pdbs = {}
        # set these to off for safety
        self.columns_viewport = 6
        self.columns_text = 6
        self.encryption = False
        self.encryption_key = str()
        self.public = Privacy.private
        self.location_viewport = Location.left
        self.freelyeditable = False
        self.new_editors = []
        self.model = False
        self.confidential = False

    def __getattr__(self, attr) -> Any:
        ## Accepts custom fields. Messes up @property, so it is done manually.
        if attr not in self.__dict__:
            return False
        else:
            return self.__dict__[attr]

    def __setattr__(self, attr, value) -> None:
        ## Accepts custom fields. Messes up @property, so it is done manually.
        dynamics = {'columns_viewport': self._set_columns_viewport, 'columns_text': self._set_columns_text}
        if attr in dynamics.keys():
            dynamics[attr].__call__(value)
        elif attr in self.preferences and not isinstance(value, self.preferences[attr]):
            # please ignore overzealous pycharm warning
            raise TypeError(f'{attr} is expected to be {self.preferences[attr]}, {value} is {type(value).__name__}')
        else:
            self.__dict__[attr] = value

    # ==== Columns =====================================================================================================

    def check_columns(self):
        """
        Verify that columns_viewport and columns_text are 12.
        See bootstrap docs for details of why a full row is 12.
        """
        if self._columns_viewport + self._columns_text != 12:
            raise ValueError('text and viewport are not 12.')

    def _get_columns_viewport(self):
        self.check_columns()
        return self.columns_viewport

    def _get_columns_text(self):
        self.check_columns()
        return self.columns_text

    def _set_columns_viewport(self, value: int):
        if not isinstance(value, int):
            raise TypeError('Expecting int')
        elif value > 12 or value < 0:
            raise ValueError('Max is 12.')
        else:
            self.__dict__['columns_text'] = 12 - value
            self.__dict__['columns_viewport'] = value

    def _set_columns_text(self, value: int):
        self.columns_viewport = 12 - value

    # ==== IO ==========================================================================================================

    def parse(self, data):
        """
        Parses server sent data.

        :return:
        """
        # proteinJSON -> protein
        # unfortunately I need to change how the data is sent. but I never get round to it.
        if isinstance(data['proteinJSON'], str):  # will be.
            self.proteins = json.loads(data['proteinJSON'])
        else:
            self.proteins = data['proteinJSON']
        del data['proteinJSON']
        if isinstance(data['pdb'], str):  # will be.
            try:
                data['pdb'] = json.loads(data['pdb'])
            except json.JSONDecodeError:
                warn('Pre-beta entry!')
                data['pdb'] = [[data['proteinJSON'][0]['value'], data['pdb']]]
        # pdb
        self.pdbs = dict(data['pdb'])
        del data['pdb']
        # date
        self.date = datetime.fromisoformat(data['date'])
        del data['date']
        self.location_viewport = Location[data['location_viewport']]
        del data['location_viewport']
        self.public = Privacy[data['public']]
        del data['public']
        # rest
        for k, v in data.items():
            setattr(self, k, v)
        return self

    def dumps(self):
        """
        For proper saving see ``to_pickle`` or ``save``.
        convert the object as dictionary.
        :return:
        """
        data = self.__dict__.copy()
        data['page'] = self.page
        data['proteinJSON'] = json.dumps(self.proteins)
        del data['proteins']
        data['pdb'] = json.dumps([[k, v] for k, v in self.pdbs.items()])
        del data['pdbs']
        del data['date']
        # enums
        data['location_viewport'] = self.location_viewport.name
        data['public'] = self.public.name
        return data

    # ======== Misc ====================================================================================================

    @property
    def link(self) -> str:
        link = f'<a href="{self.parent.url}data/{self.page}" target="_blank">{self.page}</a>' + \
               f'<img alt="error" src="{self.parent.url}thumb/{self.page}" height="200">'
        return link

    def show_link(self) -> None:
        display(HTML(self.link))

    def what_is(self, attr: str):
        """
        Query what an attribute does.

        :param attr:
        :return:
        """
        if attr in self.definitions:
            print(attr, self.definitions[attr])
        else:
            print(attr, 'No idea')

    def clear_revisions(self) -> None:
        """
        On commit, ask the server to blank all revisions. Only privedged or admin.

        :return: None
        """
        self.no_revisions = True

    def refresh_image(self) -> None:
        """
        On commit, refresh the tumbnail.

        :return: None
        """
        self.refresh_image = True

    def to_pickle(self, filename: str):
        """
        writes to pickle

        :param filename:  filename (path okay) with extension.
        :return:
        """
        pickle.dump(self.dumps(), open(filename,'wb'))

    def from_pickle(self, filename: str):
        data = pickle.load(open(filename,'wb'))
        self.page = data['page']
        self.proteins = json.loads(data['proteinJSON'])
        self.pdbs = {k: v for k, v in json.loads(data['pdb'])}
        self.location_viewport.name = data['location_viewport']
        self.public.name = data['public']

    def save(self, name: str):
        """
        Save the description, pdbs and JS for easier editing.
        It's inverse ``load`` loads them.
        The purpose is to edit the file as opposed to a wonky notebook cell.

        ``save`` is a legacy keyword in the page data, but is no longer used.
        :param name:
        :return:
        """
        with open(name + '.md', 'w') as w:
            w.write(self.description)
        with open(name + '.js', 'w') as w:
            w.write(self.loadfun)
        for pdbname in self.pdbs:
            with open(f'{name}-{pdbname}.pdbs', 'w') as w:
                w.write(self.pdbs[pdbname])

    def load(self, name: str):
        """
        Loads what was saved with ``.saved``

        :param name:  filename (path okay) with no extension.
        :return:
        """
        with open(name + '.md', 'r') as r:
            self.description = r.read()
        with open(name + '.js', 'r') as r:
            self.loadfun = r.read()
        for pdbname in self.pdbs:
            with open(f'{name}-{pdbname}.pdbs', 'r') as r:
                self.pdbs[pdbname] = r.read()

    # ======== Parent ==================================================================================================

    def rename_protein_variable(self,
                                index: Optional[int] = None,
                                name: Optional[str] = None,
                                value: Optional[str] = None,
                                newname: Optional[str] = None):
        """
        Overloaded method. accepts one of the three parameters:
        
        :param index: list index
        :param name: name of protein
        :param value: value of protein, this is the JS function
        :param newname: if None, it will remove it as a variable and change the type to data.
        :return: 
        """
        protein = self.get_protein(index, name, value)
        oldname = protein['value']
        if newname and re.search('[^\w\_]', newname) is not None:
            raise ValueError(f'{newname} is not legal')
        elif newname and 'isVariable' in protein and protein['isVariable'] in (True, 'true'):  # rename
            protein['value'] = re.sub('[^\w\_]', '', newname)
            self.pdbs[newname] = self.pdbs[oldname]
            del self.pdbs[oldname]
        elif newname:
            protein['value'] = newname
        else:
            protein['value'] = self.pdbs[oldname]
            del protein['isVariable']
            del self.pdbs[oldname]

    def get_protein(self,
                    index: Optional[int] = None,
                    name: Optional[str] = None,
                    value: Optional[str] = None) -> Dict:
        """
        Overloaded method. accepts one of the three parameters:
        
        :param index: list index
        :param name: name of protein
        :param value: value of protein, this is the JS function
        :return: the protein entry
        """
        not_none = len([v for v in (index, name, value) if v is not None])
        if not_none == 0:
            raise ValueError('Specify one of index, name, value.')
        elif not_none > 1:
            raise ValueError('Specify only one of index, name, value.')
        elif index is not None:
            return self.proteins[index]
        elif name is not None:
            return [p for p in self.proteins if 'name' in p and p['name'] == name][0]
        elif value is not None:
            return [p for p in self.proteins if 'value' in p and p['value'] == value][0]
        else:
            raise ValueError('Impossible.')

    def make_github_entry(self, username: str, repo: str, path: str, branch: str='main') -> Dict:
        """
        make a protein entry.

        :param username: Github username
        :param repo: repository
        :param path: path within repo
        :param branch: main or master etc.
        :return:
        """
        url = f'https://raw.githubusercontent.com/{username}/{repo}/{branch}/{path}'
        name = re.sub('[^\w_]', '', os.path.splitext(os.path.split(path)[1])[0])
        return {'type': 'url', 'value': url, 'name': name}

    def append_github_entry(self, username: str, repo: str, path: str) -> int:
        """
        make and add a protein entry.

        :param username: Github username
        :param repo: repository
        :param path: path within repo
        :return:
        """
        self.proteins.append(self.make_github_entry(username, repo, path))
        return len(self.proteins) - 1

    def append_pdbfile(self, filename: str, varname: Optional[str] = None, chain_definitions: List[dict] = []) -> int:
        """
        make and add a protein entry.

        :param filename: the local filename to add.
        :param varname: what should the pdbblock be known as?
        :param chain_definitions: list of dict defining who each chain is. Needed solely for Uniprot modal and chain name in builder
        :return:
        """
        pdbblock = open(filename).read()
        if varname is None:
            varname = re.sub(r'[^\w_]', '', os.path.splitext(filename)[0])
            varname = re.sub(r'^\d+', '', varname)
            print(f'Variable name is {varname}')
        self.append_pdbblock(pdbblock, varname, chain_definitions)
        return len(self.proteins) - 1

    def append_pdbblock(self, pdbblock: str, varname: str, chain_definitions: List[dict] = []) -> int:
        """
        make and add a protein entry.

        :param pdbblock: the pdbblock to add.
        :param varname: what should the pdbblock be known as?
        :param chain_definitions: list of dict defining who each chain is. Needed solely for Uniprot modal and chain name in builder
        :return:
        """
        assert varname not in self.pdbs, f'Varname {varname} already in use.'
        self.proteins.append({'type': 'data',
                              'value': varname,
                              'isVariable': 'true',
                              'chain_definitions': chain_definitions,
                              'history': {'code': '', 'changes': ''}})
        self.pdbs[varname] = pdbblock
        return len(self.proteins) - 1

    def remove_protein(self, index):
        if self.proteins[index]['type'] == 'data':
            del self.pdbs[self.proteins[index]['value']]
        d = self.proteins[index]
        del self.proteins[index]
        return d

    def add_publication(self,
                        url: Optional[str] = None,
                        authors: Optional[str] = 'TBA',
                        year: Optional[int] = None,
                        title: Optional[str] = 'TBA',
                        journal: Optional[str] = 'manuscript in preparation',
                        issue: Optional[str] = 'NA',
                        doi: Optional[str] = None):
        """
        This is an admin only operation due to the security issue (dodgy links in the gallery).
        Publication is not data from the page, it's an actual table

        :param uuid:
        :param url:
        :param authors:
        :param year:
        :param title:
        :param journal:
        :param issue:
        :return:
        """
        if year is None:
            year = datetime.year
        if url is not None:
            pass
        elif doi is None:
            url = '#'
        else:
            url = f'https://dx.doi.org/{doi}'
        return self.parent.post_json('set', {'item': 'publication',
                                             'identifier': self.page,
                                             'url': url,
                                             'title': title,
                                             'year': year,
                                             'authors': authors,
                                             'journal': journal,
                                             'issue': issue})

    # ======== Parent ==================================================================================================

    def retrieve(self):
        for k, v in self.parent.get_page(uuid=self.page).__dict__.items():
            self.__dict__[k] = v
        return self

    def commit(self):
        return self.parent.set_page(self.page, self)

    def delete(self):
        return self.parent.del_page(self.page)

    def rename(self, new_name):  ##admin only
        return self.parent.rename_page(self.page, new_name)

    def shorten(self, short_name):
        return self.parent.shorten_page(self.page, short_name)

    def __repr__(self):
        return f'<MikePage:{self.page} at {hex(id(self))}>'

    def __str__(self):
        text = f'Page {self.page}\n'
        if self.title:
            text += f'# {self.title}\n'
        if self.description and len(self.description) > 200:
            text += f'{self.description[:200]} [... +{len(self.description) - 200} chars]\n'
        elif self.description:
            text += f'{self.description}\n'
        return text


# ======== Extend docs =================================================================================================


MikePage.__doc__ += '\n Attributes:'
for key in MikePage.definitions:
    if key not in MikePage.preferences:
        MikePage.__doc__ += f'        * ``{key}``: {MikePage.definitions[key]}\n'
    else:
        MikePage.__doc__ += f'        * ``{key}``: ({MikePage.preferences[key]}) {MikePage.definitions[key]}\n'
